/*******************************************************************************
 *
 * Macros for the evaluation of scored data from Raptor.
 *
 * ****************************************
 * ****************************************
 * ********                     ***********
 * ********  SAS INTERNAL ONLY  ***********
 * ********                     ***********
 * ****************************************
 * ****************************************
 *
 * SAS Confidential and Proprietary (c) 2004
 *
 * This code is considered a BETA version --- not guaranteed to be
 * accurate.  Please use at your own discretion.
 *
 *
 * The included macros are:
 *    %evalData(in_tagged_dataset, out_evalResults,[optional keyed parameters]);
 *
 *
 * 12/24/04:
 *   Sent to client.
 *
 * 1/4/05:
 *   Resent to client after converting tagData from SAS9 to SAS8 (no
 *   change in evalData).
 *
 * 3/8/05:
 *   Please see individual macros for descriptions of changes
 *
 * 4/5/05:
 *   Please see individual macros for descriptions of changes
 *   
 * 4/18/05:
 *   Spawned off of code sent to client.
 *
 * 5/9/05:
 *   Please see evalData for a description of the change
 *
 * 7/25/07:
 *    Remove tagData from this file.
 *    Rewrite for postings support!
 *
 * 8/3/07:
 *    Allow for one-behind evaluation --- INTERNAL ONLY.
 *   
 * 3/17/09:
 *    Combine API v2 and API v3 evaluation codes into one
 *
 ******************************************************************************/



/*******************************************************************************
 * Macro to evaluate a tagged, scored data set, providing a dataset with
 * CFPR, CDR, and VDR for difference score cutoffs.
 *
 * Usage:
 *
 * %evalData (in_tagged_dataset, out_evalResults, [optional keyed parameters ]);
 *
 *
 * Inputs:
 * =======
 *   in_tagged_dataset 
 *   -----------------
 *     This is output from %tagData defined above, but now including
 *     scores.
 *
 *     SAS dataset containing a number of transactiosn with the fields:
 *       rqo_card_num          Account number
 *       smh_tran_id           Transaction-type ("10?016" = Authorization)
 *       rrr_score_1           Score
 *       rqo_proc_utc_datetime Date and Time of the transaction in UTC
 *       rqo_tag               Indicates legitimate, pre-fraud, fraud-window,...
 *
 *     For authorizations, the following fields must also be present:
 *       rca_auth_sys_dec      Decision code for the system ("A", "V", ...)
 *       rca_mod_amt/rca_amt   Amount
 *
 *     For postings, the following field must also be present:
 *       rcp_mod_amt/rcp_amt   Amount
 *       rcp_purchase_date     Purchase Date
 *
 *     MUST be sorted by: rqo_card_num rqo_proc_utc_datetime
 *
 * Output:
 * =======
 *    evalResults
 *    -----------
 *    A  dataset containing a number of observations with the following
 *    fields:
 *
 *      cutoff       The score threshold
 *      CFPR         The Account False Positive Ratio for this score cutoff
 *      CDR          The Account Detection Rate for this score cutoff
 *      VDR          The Value Detection Rate for this score cutoff
 *      numFP        The raw num of account FP for this score cutoff
 *      detFrdDol    The raw num of fraud dollars detected for this score cutoff
 *      detFrdAcct   The raw num of fraud accts detected for this score cutoff
 *
 *    And the values for the optional parameters (see below):
 *      recontact_days
 *      srate
 *      exclude_prefraud    
 *      all_tran_types
 *      INCLUDE_POSTINGS
 *      max_amount
 *      score_incr
 *
 *    For a definition of the parameters, p[lease see teh document "Flowchart
 *    for Evaluation of Raptor Fraud Detection Models", Nov 8, 2004.
 *
 *
 * Optional Parameters:
 * ====================
 *   RECONTACT_DAYS = 999,     The minimum time between contacts 
 *   SRATE = 0.10,             Sampling rate for nonfraud accounts vs.  fraud
 *                             accounts.  This is a fraction between 0 and 1. 
 *   EXCLUDE_PREFRAUD = 99999  Number of days of pre-frauds to exclude before DOFF
 *   ALL_TRAN_TYPES = 0,       Should we only consider scores on auths or everything? 
 *   INCLUDE_POSTINGS = 0,     Should we also consider scores on posts in addition to auths?
 *   ONE_BEHIND = 0,           Don't allow savings at the exact time of scoring.
 *   MAX_AMOUNT = 200000,      Don't include amounts greater than this 
 *   SCORE_INCR =  1,          The score resolution for evaluation 
 *   TMP=evalTmp,              The name to use for the temporary datasets.
 *   TRAN_FILTER=,             The code (usually macro) to run before
 *                             filtering transactions.  It may set the
 *                             flag "skipThisTran = 1".
 *   RESCALE_SCORE = N,        Rescale the score from 0-1000 to 1-999.
 *                             DO NOTE THROW AWAY 0 SCORES 
 *
 **********
 * Author: padula
 * Date: 12/22/04
 *
 * 12/24/04:
 *   Sent to client.
 *
 * 1/4/05:
 *   Resent to client after converting tagData from SAS9 to SAS8 (no
 *   change in evalData).
 *
 * 3/8/05:
 *     - Added TRAN_LEVEL_VDR option.
 *     - Fixed bug giving incorrect weight to prefraud transactions.
 *
 * 4/5/05:
 *     - Removed TRAN_LEVEL_VDR option.
 *     - Changed field names to coincide with latest API
 *       + Includes using rqo_card_num for the account number
 *       + Includes using rqo_proc_utc_datetime for the processing time:
 *         tran_date/tran_time have been redefined as the "effective"
 *         date, not the processing date.  [Note: for auths, these are the
 *         same.  For non-mons and payments, these may differ
 *         significantly.]
 *     - Added optional parameter to allow user to specify the temporary
 *       dataset's name.
 *     - Updated default values of parameters to reflect expected
 *       evaluation:
 *          RECONTACT_DAYS = 7   -> 999
 *          INCLUDE_PREFRAUD = 1 ->  0
 *          MAX_AMOUNT = 50000   -> 200000
 *
 * 4/18/05:
 *     - Added ability to call a filter.
 *
 * 5/9/05:
 *   Approved dollars on auths must now have a decision of "A" ---
 *   removed "V" from consideration.
 *
 * 7/25/07:
 *    Rewrite for postings support!
 * 
 * TODO :
 *   We do not account for postings
 *
 * 3/19/09: 
 *   1) Changed the INCLUDE_PREFRAUD switch to EXCLUDE_PREFRAUD
 *      This now tells the number of days of prefraud to ignore prior 
 *      to the DOFF.
 *   2) Added number of skipped transactions to the output
 *   3) Added # of monthly non-fraud and fraud cards to the output
 *   4) Added Mean/Median time to detect to the output
 *   5) Added the OSF: Outsort Factor to the output
 *   6) Code will fail when a card is not always tagged as '0' or non-'0'
 *   7) Code supports multiple included filters, just pass the list in quotes
 *      i.e. -f "filter1.inc filter2.inc".  This code will add them in sequentially.
 *   
 ******************************************************************************/

%macro evalData (
    in_tagged_dataset,        /* REQUIRED */
    evalResults,              /* REQUIRED */

    RECONTACT_DAYS = 999,     /* The minimum time between contacts */
    SRATE = 0.10,             /* Sampling rate for nonfraud accounts vs.  fraud
                                 accounts.  This is a fraction between 0 and 1. */
    EXCLUDE_PREFRAUD = 99999, /* Number of days of pre-frauds to exclude before DOFF */
    ALL_TRAN_TYPES = 0,       /* Should we only consider scores on auths or everything? */
    INCLUDE_POSTINGS = 0,     /* Should we also consider scores on posts in addition to auths? */
    ONE_BEHIND = 0,           /* Don't allow savings at the exact time of scoring. */
    MAX_AMOUNT = 200000,      /* Don't include amounts greater than this */
    SCORE_INCR =  1,          /* The score resolution for evaluation */
    TMP=evalTmp,              /* The name to use for the temporary datasets. */
    TRAN_FILTER=,             /* The code (usually macro) to run before
                                 filtering transactions. */
    RESCALE_SCORE = N,        /* Rescale the score from 0-1000 to 1-999.
                                 DO NOTE THROW AWAY 0 SCORES */
);

%local filter i;
%let i=1;

data &TMP.;
  /* Determine the number of cuttof scores we will have given the max score of
   * 1000 and an increment of SCORE_INC.  Save this in a macro variable so that
   * we may use it in array declarations.  Also, eliminate unwanted spaces by
   * explicitly converting to a string and trim'ing before we perform a symput
   * call.
   */
  call symput ('N_CUTOFF', trim(left(put(ceil(1000 / &SCORE_INCR) + 1, 8.))));
run;


/*******************************************************************************
 *
 * Read in all transations and evaluate according to the labels:
 *
 *    0 -- "non-fraud": an account which is not compromised at any time during
 *         the evaluation period.
 *
 *    1 -- "fraud": denotes that this transaction occured while an account is in
 *          a state of compromised
 *
 *    2 -- "pre-fraud": denotes a transaction on an account that will be
 *         compromised at a later time during this evaluation period (this is
 *         different for sampling purposes)
 *
 *    3 -- "post-status": a transaction on a compromised account after the
 *         account has been blocked.  These are eliminated from the evaluation.
 *
 *
 * TAG             Description         Catchable  Savable  Loss   Eval
 * 
 * Authorizations:
 *
 *  1              fraud                  Y          Y       Y      Y
 *  2              pre-fraud              N          N       N      N
 *  3              post-fraud             N          N       Y      N
 *
 * Postings:
 *
 *  A              Purch Time <  DOFF     N          N       N      N
 *                 Proc  Time <  FFT
 *  
 *  B              Purch Time <  DOFF     Y          N       N      Y
 *         FFT  <= Proc  Time <= LAFA
 *  
 *  C              Purch Time <  DOFF     N          N       N      N
 *                 Proc  Time >  LAFA
 *  
 *  D      DOFF <= Purch Time <  FFT      Y          N       Y      Y
 *         FFT  <= Proc  Time <= LAFA
 *  
 *  E      DOFF <= Purch Time <  FFT      N          N       Y      Y
 *                 Proc  Time >  LAFA
 *  
 *  F      FFT  <= Purch Time <= LAFA     Y          Y       Y      Y
 *         FFT  <= Proc  Time <= LAFA
 *  
 *  G      FFT  <= Purch Time <= LAFA     N          Y       Y      Y
 *                 Proc  Time >  LAFA
 *  
 *  H              Purch Time >  LAFA     N          N       Y      Y
 *                 Proc  Time >  LAFA
 *
 *
 ******************************************************************************/
data _null_;
  startdate="&TSTART."d;
  enddate="&TEND."d;
  evalmonths=1+intck('month',startdate,enddate);
  call symput('eval_startdate',startdate);
  call symput('eval_enddate',enddate);
  call symput('num_eval_months',evalmonths);
run;
%let eval_startdate=%cmpres(&eval_startdate);
%let eval_enddate=%cmpres(&eval_enddate);
%let num_eval_months=%cmpres(&num_eval_months);

%let cardleveldatafields=&CARDNUMBER FRD_FIRST_DATE_RES acct_CouldBeDetected has_tag_0 n_skipped n_skipped_nonDate
     acct_days_to_detect_: acct_trxn_to_detect_: acct_alerts_: acct_has_alert_: acct_skipped_p acct_skipped_nonDate_p
     acct_trxns frd_yymm;

data  &TMP.(drop=&cardleveldatafields)
      cardleveldata(keep=&cardleveldatafields)
      ;
  set &in_tagged_dataset END=lastTransaction;
  by &CARDNUMBER rqo_proc_utc_datetime;

  if _n_=1 then do; 
    length yymm yymm_nf_count 8;
    declare hash hfar(hashexp:5);
    hfar.defineKey('yymm');
    hfar.defineData('yymm','yymm_nf_count');
    hfar.defineDone();
    call missing(yymm,yymm_nf_count);
  end;

  /*****************************************************************************
   *
   * Declare and initialize global variables
   *
   ****************************************************************************/

  array cutoffs           {&N_CUTOFF} cutoffs1-cutoffs&N_CUTOFF.;
  array global_numFP      {&N_CUTOFF} gbl_numFP1-gbl_numFP&N_CUTOFF.;
  array global_dolFP      {&N_CUTOFF} gbl_dolFP1-gbl_dolFP&N_CUTOFF.;
  array global_detFrdDol  {&N_CUTOFF} gbl_detFrdDol1-gbl_detFrdDol&N_CUTOFF.;
  array global_detFrdAcct {&N_CUTOFF} gbl_detFrdAcct1-gbl_detFrdAcct&N_CUTOFF.;
  n_cutoffs = &N_CUTOFF;

  retain
    cutoffs                     /* The cutoffs we will be examining        */
    global_numFP                /* The number of false positives              */
    global_dolFP                /* The number of false positives              */
    global_totLegitDol          /* The total legitimate approved dollars      */
    global_detFrdDol            /* The approved fraud dollars after detection */
    global_detFrdAcct           /* The number of fraud accts detected         */
    global_totFrdDol            /* The total approved fraud dollars           */
    global_totFrdAcct           /* The total number of fraud accounts         */
    n_cutoffs                   /* The number of cutoffs                   */
  ;
  if (_N_ = 1) then do;
    do i=1 to n_cutoffs;
        cutoffs{i} = (i - 1)* &SCORE_INCR;
        global_numFP{i} = 0;
        global_dolFP{i} = 0;
        global_detFrdDol{i} = 0;
        global_detFrdAcct{i} = 0;
    end;
    global_totLegitDol = 0;
    global_totFrdDol = 0;
    global_totFrdAcct = 0;
  end;

  /*************************************
   *
   * Variables to keep in the &TMP 
   * dataset.
   *
   ************************************/
  keep
    cutoffs1-cutoffs&N_CUTOFF
    gbl_numFP1-gbl_numFP&N_CUTOFF
    gbl_dolFP1-gbl_dolFP&N_CUTOFF
    gbl_detFrdDol1-gbl_detFrdDol&N_CUTOFF
    gbl_detFrdAcct1-gbl_detFrdAcct&N_CUTOFF
    global_totLegitDol
    global_totFrdDol
    global_totFrdAcct
    n_cutoffs
  ;

  /*************************************
   *
   * Variables to keep in the cardleveldata
   * dataset.
   *
   ************************************/
  keep
    &cardLeveldatafields;
  ;

  /*************************************
   *
   * Set the weighting due to sampling.
   * We assume that compromised accounts
   *   are presented in toto, while
   *   non-fraud accounts have been
   *   sampled *At The Account Level*.
   *
   ************************************/
  legit_acct_weight = 1./&SRATE;
  fraud_acct_weight = 1;

  /*****************************************************************************
   *
   * Declare account-level variables
   *
   ****************************************************************************/
  array acct_lastContact  {&N_CUTOFF};
  array acct_isDetected   {&N_CUTOFF};
  array acct_whenDetected {&N_CUTOFF};
  array counted_{&num_eval_months} counted_1-counted_&num_eval_months.;
  array acct_days_to_detect_{&N_CUTOFF} acct_days_to_detect_1-acct_days_to_detect_&N_CUTOFF;
  array acct_trxn_to_detect_{&N_CUTOFF} acct_trxn_to_detect_1-acct_trxn_to_detect_&N_CUTOFF;
  array acct_alerts_{&N_CUTOFF} acct_alerts_1-acct_alerts_&N_CUTOFF;
  array acct_has_alert_{&N_CUTOFF} acct_has_alert_1-acct_has_alert_&N_CUTOFF;
  length first_tag $1;
  retain
    acct_lastContact                    /* Date this acct was last contacted */
    acct_isDetected                     /* Has this acct been detected */
    acct_whenDetected                   /* When this acct was detected */
    acct_couldBeDetected                /* Has this account had at least one
                                           transaction after compromise and
                                           before block */
    has_tag_0
    counted_1-counted_&num_eval_months.
    acct_days_to_detect_1-acct_days_to_detect_&N_CUTOFF
    acct_trxn_to_detect_1-acct_trxn_to_detect_&N_CUTOFF
    n_skipped
    n_skipped_nonDate
    acct_trxns
    acct_frd_trxns
    acct_alerts_1-acct_alerts_&N_CUTOFF
    acct_has_alert_1-acct_has_alert_&N_CUTOFF
    first_tag
  ;

  if first.&CARDNUMBER then do;
    do i=1 to n_cutoffs;
      acct_lastContact{i}     = "01Jan1960"d;
      acct_isDetected{i}      = 0;
      acct_whenDetected{i}    = 0;
      call missing(acct_days_to_detect_{i},acct_trxn_to_detect_{i});
      acct_alerts_{i}         = 0;
      acct_has_alert_{i}      = 0;
    end;
    acct_couldBeDetected = 0;
    has_tag_0 = 0;
    do i=1 to &num_eval_months;
      counted_{i}=0;
    end;
    n_skipped=0;
    n_skipped_nonDate=0;
    acct_trxns = 0;
    acct_frd_trxns = 0;
    call missing(first_tag);
  end;

  /*****************************************************************************
   *
   * Filter out invalid transactions!
   *
   ****************************************************************************/
  skipThisTran = 0;
  %if "&TRAN_FILTER" ne "" %then %do;
    /*%include "&TRAN_FILTER.";*/
    %do %until(%quote(%scan(%quote(&TRAN_FILTER),&i,%str( )%str(,)))=%str());
      %let filter=%scan(%quote(&TRAN_FILTER),&i,%str( )%str(,));
      %include "&filter";
      %let i=%eval(&i.+1);
    %end;
  %end;

  %if &RESCALE_SCORE=Y %then %do;
    rrr_score_1 = round((rrr_score_1 * (499/500)) + 1);
    rrr_score_1 = min(rrr_score_1, 999);
    rrr_score_1 = max(rrr_score_1,   1);
  %end;
  if rrr_score_1 = 0        then skipThisTran = 1;

  /* Missing is a large negative value */
  if &AUTH_AMT > &MAX_AMOUNT  then skipThisTran = 1;

  %if &INCLUDE_POSTINGS=1 %then %do;
    if &POST_AMT > &MAX_AMOUNT  then skipThisTran = 1;
  
    %if &APIVERSION = 2 OR &APIVERSION = 3 %then %do;
      if (&TRAN_ID in ("100007","101007","100019","101019") and missing(&POST_PURCH_DATE.)) then skipThisTran = 1;
    %end;
    %else %do;
      if (&TRAN_ID in ("CP") and missing(&POST_PURCH_DATE.)) then skipThisTran = 1;
    %end;
  %end;

  skipThisTran_preDates=skipThisTran;
  if &TAG. ne '0' and FRD_FIRST_DATE_RES < &EVAL_STARTDATE. then skipThisTran=1;
  if rqo_proc_utc_datetime < 86400* &EVAL_STARTDATE. or rqo_proc_utc_datetime >= 86400 * (&EVAL_ENDDATE. + 1) then skipThisTran=1;

  %if &ALL_TRAN_TYPES = 1 %then %do;
    validTranID=1;
  %end;
  %else %if &INCLUDE_POSTINGS = 1 %then %do;
    %if &APIVERSION = 2 OR &APIVERSION = 3 %then %do;
      if &TRAN_ID in ('100000','101000','100007','101007','100016','100019','101016','101019') then validTranID=1;
    %end;
    %else %do;
      if &TRAN_ID in ("CA","CP") then validTranID=1;
    %end;
    else validTranID=0;
  %end;
  %else %do;
    %if &APIVERSION = 2 OR &APIVERSION = 3 %then %do;
      if &TRAN_ID in ('100000','101000','100016','101016') then validTranID=1;
    %end;
    %else %do;
      if &TRAN_ID in ("CA") then validTranID=1;
    %end;
    else validTranID=0;
  %end;

  /*********************
   * 
   * Count the unique non-fraud cards that appear each month
   *
   ********************/
  yymm=year(datepart(rqo_proc_utc_datetime))*100+month(datepart(rqo_proc_utc_datetime));
  if skipThisTran=0 and validTranID=1 and ((&TAG. = "0") or 
     (&TAG. in ("2","A") and datepart(rqo_proc_utc_datetime) <= FRD_FIRST_DATE_RES - &EXCLUDE_PREFRAUD)) then do;
    eval_index=12*(year(datepart(rqo_proc_utc_datetime))-year(&eval_startdate)) +
               month(datepart(rqo_proc_utc_datetime)) - month(&eval_startdate) + 1;;
    if counted_{eval_index}=0 then do;
      rc=hfar.find();
      if rc ne 0 then do;
        if &TAG.="0" then yymm_nf_count=legit_acct_weight;
        else yymm_nf_count=fraud_acct_weight;
        yymm=year(datepart(rqo_proc_utc_datetime))*100+month(datepart(rqo_proc_utc_datetime));
        rc=hfar.add();
      end;
      else do;
        if &TAG.="0" then yymm_nf_count=yymm_nf_count+legit_acct_weight;
        else yymm_nf_count=yymm_nf_count+fraud_acct_weight;
        rc=hfar.replace();
      end;
      counted_{eval_index}=1;
    end;
  end;

  /*****************************************************************************
   *
   * Check to make sure all tags for a particular card are either all '0' or all 
   * non-'0'.
   *
   ****************************************************************************/
  if validTranID=1 and missing(first_tag) then first_tag=&TAG;
  if skipThisTran=0 and validTranID=1 and ((&TAG='0' and first_tag ne '0') OR (&TAG ne '0' and first_tag = '0')) then do;
    putlog "RMD ERROR!!!!! Card has both 0 and non-0 tags: " &CARDNUMBER= first_tag= &TAG.=;
    abort return;
  end;

  /*************************************
   *
   * Count up the number of transactions
   * skipped for each card
   *
   ************************************/
   if validTranID=1 then do;
     if skipThisTran = 1 then n_skipped = n_skipped + 1;
     else acct_trxns = acct_trxns + 1;
     if skipThisTran_preDates = 1 then n_skipped_nonDate = n_skipped_nonDate + 1;
   end;

  /*************************************
   *
   * Examine the transactions on this 
   * acct that have "valid" scores.
   *
   ************************************/
  if skipThisTran = 0 and validTranID = 1 then do;

      my_index = floor(rrr_score_1/&SCORE_INCR) + 1;

      if (&TAG. = "0") or 
         (&TAG. in ("2","A") and datepart(rqo_proc_utc_datetime) <= FRD_FIRST_DATE_RES - &EXCLUDE_PREFRAUD) then do;

        %if &APIVERSION = 2 OR &APIVERSION = 3 %then %do;
          if (&TRAN_ID in ("100000","101000","100016","101016") and &AUTH_DEC. in ("A","1")) then do;
        %end;
        %else %do;
          if (&TRAN_ID = "CA" and &AUTH_DEC in ("A","1")) then do;
        %end;
          appDol = &AUTH_AMT.;
        end;
        %if &INCLUDE_POSTINGS=1 %then %do;
          %if &APIVERSION = 2 OR &APIVERSION = 3 %then %do;
            else if (&TRAN_ID in ("100007","101007","100019","101019")) then do;
          %end;
          %else %do;
            else if (&TRAN_ID = "CP") then do;
          %end;
            appDol = &POST_AMT.;
          end;
        %end;
        else do;
          appDol = 0;
        end;

        if (&TAG. = "0") then do;
          global_totLegitDol = global_totLegitDol + legit_acct_weight*appDol;
        end;
        else if (&TAG. in ("2","A")) then do;
          global_totLegitDol = global_totLegitDol + fraud_acct_weight*appDol;
        end;

        do i=1 to my_index;                                     /* non-fraud */
          if ((rrr_score_1 >= cutoffs{i}) and
            (datepart(rqo_proc_utc_datetime) >= acct_lastContact{i} + &RECONTACT_DAYS)) then do;
              acct_alerts_{i}=acct_alerts_{i} + 1;
              acct_has_alert_{i}=1;
              if (&TAG. = "0") then do;
                global_numFP{i} = global_numFP{i} + legit_acct_weight;
                global_dolFP{i} = global_dolFP{i} + legit_acct_weight*appDol;
              end;
              else if (&TAG. in ("2","A")) then do;
                global_numFP{i} = global_numFP{i} + fraud_acct_weight;
                global_dolFP{i} = global_dolFP{i} + fraud_acct_weight*appDol;
              end;
              acct_lastContact{i} = datepart(rqo_proc_utc_datetime);
            end;
        end;                                                     /* non-fraud */

        if (&TAG. = "0") then has_tag_0 = 1;

      end;
      else if (&TAG. in ("1", "B", "D", "F", "G")) then do; /* fraud compromise window */

          if &TAG. in ("1","B", "D", "F") then
            is_catchable = 1;
          else
            is_catchable = 0;

          if &TAG. in ("1","F", "G") then
            is_saveable = 1;    /* Hallelujah! */
          else
            is_saveable = 0;    /* Abandon all hope, ye who enter here */

          if is_catchable = 1 then do;
            /***********
             * We have a valid score in the compromise window!
             * Determine if we have a valid approved or verified amount.
             ***********/
            acct_couldBeDetected = 1;

            /**********
             * Look at each cutoff independently.
             * As we may be examining pre-fraud for false-positives, each
             * cutoff will have it's own re-contact issues, and therefore we
             * cannot just sum up or down.
             *********/
            do i=1 to n_cutoffs;                           /* catchable fraud cutoff loop */
              /***********
               * If we haven't been detected previously, can we do so now?
               ***********/
               if (acct_isDetected{i} = 0) then do;
                 if ((rrr_score_1 >= cutoffs{i}) and
                  (datepart(rqo_proc_utc_datetime) >= acct_lastContact{i} + &RECONTACT_DAYS)) then do;
                    acct_days_to_detect_{i} = datepart(rqo_proc_utc_datetime) - FRD_FIRST_DATE_RES;
                    acct_trxn_to_detect_{i} = acct_frd_trxns;
                    acct_isDetected{i} = 1;
                    acct_whenDetected{i} = rqo_proc_utc_datetime;
                    if "&ONE_BEHIND" = "1" then acct_whenDetected{i} = acct_whenDetected{i} + 0.01;
                 end;
               end;
            end;                                           /* catchable fraud cutoff loop */
            acct_frd_trxns = acct_frd_trxns + 1;
          end;
         
          if is_saveable = 1 then do;
            %if &APIVERSION = 2 OR &APIVERSION = 3 %then %do;
              if (&TRAN_ID in ("100000","101000","100016","101016") and &AUTH_DEC. in ("A","1")) then do;
            %end;
            %else %do;
              if (&TRAN_ID = "CA" and &AUTH_DEC in ("A","1")) then do;
            %end;
              appdol = &AUTH_AMT.;
              purch_date = rqo_proc_utc_datetime;
            end;
            %if &INCLUDE_POSTINGS=1 %then %do;
              %if &APIVERSION = 2 OR &APIVERSION = 3 %then %do;
                else if (&TRAN_ID in ("100007","101007","100019","101019")) then do;
              %end;
              %else %do;
                else if (&TRAN_ID = "CP") then do;
              %end;
                appdol = &POST_AMT.;
                /* Ignore time of day for postings as they need to have
                 * been detected on the previous day to have any effect.
                 */
                purch_date = (&POST_PURCH_DATE. * 24 + 12)*60*60;
              end;
            %end;
            else do;
              appdol = 0;
            end;
            
            if appdol > 0 then do;
              global_totFrdDol = global_totFrdDol + appDol * fraud_acct_weight;

              do i=1 to n_cutoffs;                           /* saveable fraud cutoff loop */
                /***********
                 * Sum the dollars detected
                 ***********/
                if (acct_isDetected{i} = 1 and acct_whenDetected{i} <= purch_date) then do;
                    global_detFrdDol{i} = global_detFrdDol{i} + appDol * fraud_acct_weight;
                end;
              end;                                           /* saveable fraud cutoff loop */
            end;
          end;
        end;                                       /* fraud compromise window */
    end;                                           /* Is this a valid score ? */

    if last.&CARDNUMBER then do;
      if acct_couldBeDetected = 1 then
        global_totFrdAcct = global_totFrdAcct + fraud_acct_weight;

      do i=1 to n_cutoffs;
        if acct_isDetected{i} = 1 then
          global_detFrdAcct{i} = global_detFrdAcct{i} + fraud_acct_weight;
      end;
      if (n_skipped+acct_trxns>0) then do;
        acct_skipped_p=n_skipped/(n_skipped+acct_trxns);
        acct_skipped_nonDate_p=n_skipped_nonDate/(n_skipped+acct_trxns);
      end;
      else call missing(acct_skipped_p,acct_skipped_nonDate_p);
      frd_yymm=year(FRD_FIRST_DATE_RES)*100+month(FRD_FIRST_DATE_RES);
        if 0<=acct_skipped_p<1 then output cardleveldata;
    end;


    /***********************************
     *
     * Output if this is the last transaction!
     *
     **********************************/
    if lastTransaction then do;
      output &TMP.;
      rc=hfar.output(dataset:'monthlyData');
    end;
run;

data &evalResults;
  set &TMP.;

  array cutoffs           {&N_CUTOFF} cutoffs1-cutoffs&N_CUTOFF;
  array global_numFP      {&N_CUTOFF} gbl_numFP1-gbl_numFP&N_CUTOFF;
  array global_dolFP      {&N_CUTOFF} gbl_dolFP1-gbl_dolFP&N_CUTOFF;
  array global_detFrdDol  {&N_CUTOFF} gbl_detFrdDol1-gbl_detFrdDol&N_CUTOFF;
  array global_detFrdAcct {&N_CUTOFF} gbl_detFrdAcct1-gbl_detFrdAcct&N_CUTOFF;
  
  recontact_days   = &RECONTACT_DAYS;
  srate            = &SRATE;
  exclude_prefraud = &EXCLUDE_PREFRAUD;
  all_tran_types   = &ALL_TRAN_TYPES;
  include_postings = &INCLUDE_POSTINGS;
  one_behind       = &ONE_BEHIND;
  max_amount       = &MAX_AMOUNT;
  score_incr       = &SCORE_INCR;

  totFrdDol        = global_totFrdDol;
  totFrdAcct       = global_totFrdAcct;
  totLgtDol        = global_totLegitDol;

  /*************************************
   *
   * Count down from high score to low,
   * outputting our data along the way.
   *
   ************************************/
  do i=&N_CUTOFF to 1 by -1;
    cutoff         = cutoffs{i};
    numFP          = global_numFP{i};
    dolFP          = global_dolFP{i};
    detFrdDol      = global_detFrdDol{i};
    detFrdAcct     = global_detFrdAcct{i};

    if detFrdAcct > 0 then &FalsePosRate. = numFP / detFrdAcct;
    else                   &FalsePosRate. = 0;
   
    /*
    if detFrdDol  > 0 then VFPR = dolFP / detFrdDol;
    else                   VFPR = 0;
    */
   
    if totFrdAcct > 0 then &DetectRate. = 100 * detFrdAcct / totFrdAcct;
    else                   &DetectRate. = 0;
                             
    if totFrdDol  > 0 then VDR = 100 * detFrdDol / totFrdDol;
    else                   VDR = 0;

    if cutoffs{1} ^= 0     then putlog "ERROR: cutoffs{1} ^= 0 " cutoffs{1}=;
    if global_numFP{1} > 0 then &ReviewRate. = 100 * global_numFP{i} / global_numFP{1};
    else                        &ReviewRate. = 0;

    if global_numFP{1} + global_detFrdAcct{1} > 0 then
      Outsort = (global_numFP{i} + global_detFrdAcct{i}) / (global_numFP{1} + global_detFrdAcct{1});
    else
      Outsort = 0;
        
    OSF = (&FalsePosRate.+1) * &DetectRate.;

    /*
    if global_dolFP{1} > 0 then LVR = 100 * global_dolFP{i} / totLgtDol;
    else                        LVR = 0;
    */
                             
    keep 
      cutoff
      numFP
      dolFP
      detFrdDol
      detFrdAcct
      &FalsePosRate.
      /*VFPR*/
      &DetectRate.
      VDR
      &ReviewRate.
      /*LVR*/
      Outsort    
      OSF
      totFrdDol
      totLgtDol
      recontact_days
      srate
      exclude_prefraud
      all_tran_types
      include_postings
      one_behind
      max_amount
      score_incr
    ;
    output;
  end;
run;

proc means data=cardleveldata noprint;
  var n_skipped n_skipped_nonDate acct_trxns acct_skipped_p acct_skipped_nonDate_p;
  class has_tag_0;
  output out=skip_info mean()= p50()= /autoname;
run;
data skip_info;
  set skip_info;
  length Type $10;
  if has_tag_0=1 then type='Non-Fraud';
  else type='Fraud';
  if not missing(has_tag_0);
  drop has_tag_0 _type_ _freq_;
run;

proc freq data=cardleveldata(where=(has_tag_0=0 and acct_couldBeDetected=1)) noprint;
  tables frd_yymm/out=doff_info;
run;
proc sort data=doff_info;
  by frd_yymm;
run;
proc sort data=monthlyData;
  by yymm;
run;
proc means data=doff_info noprint;
  var count;
  output out=doff_info2 mean(count)=f_avg;
run;
%global f_avg;
data _null_;
  set doff_info2;
  call symput('f_avg',f_avg);
run;

data cardleveldata;
  set cardleveldata;
  array acct_days_to_detect_{&N_CUTOFF} acct_days_to_detect_1-acct_days_to_detect_&N_CUTOFF;
  array acct_trxn_to_detect_{&N_CUTOFF} acct_trxn_to_detect_1-acct_trxn_to_detect_&N_CUTOFF;
  array acct_alerts_{&N_CUTOFF} acct_alerts_1-acct_alerts_&N_CUTOFF;
  array acct_has_alert_{&N_CUTOFF} acct_has_alert_1-acct_has_alert_&N_CUTOFF;
  keep cutoff has_tag_0 acct_couldbedetected;
  keep acct_days_to_detect acct_trxn_to_detect acct_alerts acct_has_alert;
  do i = &N_CUTOFF to 1 by -1;
    cutoff = (i-1) * &SCORE_INCR;
    acct_days_to_detect=acct_days_to_detect_{i};
    acct_trxn_to_detect=acct_trxn_to_detect_{i};
    acct_alerts=acct_alerts_{i};
    acct_has_alert=acct_has_alert_{i};
    output;
  end;
run;
proc means data=cardleveldata(where=(has_tag_0=1)) noprint;
  var acct_alerts acct_has_alert;
  class cutoff;
  output out=alert_info mean()= p50()= /autoname;
run;
proc sort data=alert_info(where=(not missing(cutoff)));
  by descending cutoff;
run;
proc means data=cardleveldata(where=(has_tag_0=0 and acct_couldBeDetected=1)) noprint;
  var acct_days_to_detect acct_trxn_to_detect;
  class cutoff;
  output out=detect_info mean()= p50()= /autoname;
run;
proc sort data=detect_info(where=(not missing(cutoff)));
  by descending cutoff;
run;
proc datasets library=work noprint nowarn;
  delete cardleveldata;
run;

%global nf_avg;
proc means data=monthlyData noprint;
  var yymm_nf_count;
  output out=actinfo mean(yymm_nf_Count)=nf_avg;
run;
data _null_;
  set actinfo;
  call symput('nf_avg',nf_avg);
run;

data &evalResults;
  merge &evalResults(in=a)
        alert_info(in=b keep=cutoff acct_alerts_: acct_has_alert_:)
        detect_info(in=c keep=cutoff acct_days_to_detect_: acct_trxn_to_detect_:)
        ;
  by descending cutoff;
  drop acct_alerts_: acct_has_alert_: acct_days_to_detect_: acct_trxn_to_detect_:;
  APNF_Mean=acct_alerts_mean;
  APNF_Med=acct_alerts_p50;
  PNFA_Mean=acct_has_alert_mean;
  PNFA_Med=acct_has_alert_p50;
  DTD_Mean=acct_days_to_detect_mean;
  DTD_Med=acct_days_to_detect_p50;
  TTD_Mean=acct_trxn_to_detect_mean;
  TTD_Med=acct_trxn_to_detect_p50;
  NF_CARDS_PER_MONTH=&nf_avg;
  DOFF_PER_MONTH=&f_avg;
run;


%mend evalData;
/******************************************************************************/
