/*
 * Compute simple population statistics from a transaction dataset 
 * 
 * Author: Song, Qiang <qiang.song@sas.com> 2015
 * 
 */

/*
 * The input transaction dataset contains transaction from marked Entitys and cleas
 * which are sampled at different rate at Entity level, which is indicated with wts_var
 *
 * In miminal, the following fields should be available:
 * rqo_tran_date, rqo_proc_utc_datetime, tca_auth_sys_dec, tca_mod_amt, afd_frd_tag
 * afd_match_frd_txn,   
 *
 */


%macro get_population_statistics(in = , out = );
proc sql noprint;
    create view _tmp_stats_ as
        select
            min(rqo_tran_date) as begin_date label = "Evaluation Begin Date",
            max(rqo_tran_date) as end_date label = "Evaluation End Date",

            avg(wts_var) as wts_var,

            case afd_frd_tag
                when '0' then 'N'
                else 'Y'
            end as frd_ind,
        
            count(distinct &entity) as cards_cnt label = "Total Entity Count",

            count(*) as trans_cnt label = "Total Transaction Count",
        
            sum(tca_auth_sys_dec = '1') as approved_trans_cnt label = "Total Approved Trans Count",

            sum((tca_auth_sys_dec = '1') * (afd_frd_tag = '1') * (afd_match_frd_txn = 'M'))
                as approved_frd_trans_cnt label = "Total Approved Frd Trans",

            sum(tca_mod_amt * (tca_auth_sys_dec = '1')) as approved_amt label = "Total Approved Amount",

            sum(tca_mod_amt * (tca_auth_sys_dec = '1') * (afd_frd_tag = '1') * (afd_match_frd_txn = 'M'))
                as approved_frd_amt label = "Total Approved Frd Amount",
        
            calculated approved_trans_cnt / calculated trans_cnt as trans_approval_rate label = "Trans Approval Rate",

            calculated trans_cnt / calculated cards_cnt as trans_cnt_per_card label = "Avg Trans Cnt per Entity" 
        from &in
        group by calculated frd_ind;

    
    create table &out as
        select
            min(begin_date) as begin_date label = "Evaluation Begin Date",
            max(end_date) as end_date label = "Evaluation End Date",
            intck('month', calculated begin_date, calculated end_date) + 1 as num_months label = "Num of Months",
        

            sum(wts_var * cards_cnt) as cards_cnt label = "Total Entity Count",

            sum(wts_var * trans_cnt) as trans_cnt label = "Total Transaction Count",

            sum(wts_var * approved_trans_cnt) as approved_trans_cnt label = "Total Approved Trans Count",

            sum(wts_var * approved_amt) as approved_amt label = "Total Approved Amount",
        
            calculated trans_cnt / calculated cards_cnt as trans_cnt_per_card label = "Avg Trans Cnt per Entity",

            calculated approved_trans_cnt / calculated trans_cnt as trans_approval_rate label = "Trans Approval Rate",

            sum((wts_var * cards_cnt)*(frd_ind = 'Y')) / sum(wts_var * cards_cnt) / calculated num_months as normalized_case_frd_rate
               label = "Monthly Case Fraud Rate",
        
            sum(wts_var * approved_frd_trans_cnt) / sum(wts_var * approved_trans_cnt) as trans_frd_rate
               label = "Trans Fraud Rate",

            sum((wts_var * approved_frd_amt)*(frd_ind = 'Y')) / sum(wts_var * approved_amt) as value_frd_rate
               label = "Value Fraud Rate"
        from _tmp_stats_;

    
quit;        


%mend get_population_statistics;
    
